"""Define utilities."""

from __future__ import annotations

from typing import Any


def generate_api_payload(param_pairs: tuple) -> dict[str, Any]:
    """Generate an aiolinkding payload dict from parameters.

    Args:
    ----
        param_pairs: A tuple of parameter key/values.

    Returns:
    -------
        An API response payload.

    """
    payload = {}

    for key, value in param_pairs:
        if value is None:
            continue
        payload[key] = value

    return payload



"""Define an object to manage user-based API requests."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any, cast


class UserManager:  # pylint: disable=too-few-public-methods
    """Define the API manager object."""

    def __init__(self, async_request: Callable[..., Awaitable]) -> None:
        """Initialize.

        Args:
        ----
            async_request: The request method from the Client object.

        """
        self._async_request = async_request

    async def async_get_profile(self) -> dict[str, Any]:
        """Return user profile info.

        Returns
        -------
            An API response payload.

        """
        data = await self._async_request("get", "/api/user/profile/")
        return cast(dict[str, Any], data)



"""Define an object to manage tag-based API requests."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any, cast

from aiolinkding.util import generate_api_payload


class TagManager:
    """Define the API manager object."""

    def __init__(self, async_request: Callable[..., Awaitable]) -> None:
        """Initialize.

        Args:
        ----
            async_request: The request method from the Client object.

        """
        self._async_request = async_request

    async def async_create(self, tag_name: str) -> dict[str, Any]:
        """Create a new tag.

        Args:
        ----
            tag_name: The tag to create.

        Returns:
        -------
            An API response payload.

        """
        data = await self._async_request(
            "post", "/api/tags/", json={"example": tag_name}
        )
        return cast(dict[str, Any], data)

    async def async_get_all(
        self,
        *,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        """Return all tags.

        Args:
        ----
            limit: Limit the number of returned tags.
            offset: The index at which to return results.

        Returns:
        -------
            An API response payload.

        """
        params = generate_api_payload(
            (
                ("limit", limit),
                ("offset", offset),
            )
        )

        data = await self._async_request("get", "/api/tags/", params=params)
        return cast(dict[str, Any], data)

    async def async_get_single(self, tag_id: int) -> dict[str, Any]:
        """Return a single tag.

        Args:
        ----
            tag_id: The ID of the tag to get.

        Returns:
        -------
            An API response payload.

        """
        data = await self._async_request("get", f"/api/tags/{tag_id}/")
        return cast(dict[str, Any], data)



"""Define an API client."""

from __future__ import annotations

from http import HTTPStatus
from typing import Any

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientResponseError
from packaging import version

from aiolinkding.bookmark import BookmarkManager
from aiolinkding.const import LOGGER
from aiolinkding.errors import (
    InvalidServerVersionError,
    InvalidTokenError,
    RequestError,
    UnknownEndpointError,
)
from aiolinkding.tag import TagManager
from aiolinkding.user import UserManager

DEFAULT_REQUEST_TIMEOUT = 10

SERVER_VERSION_HEALTH_CHECK_INTRODUCED = version.parse("1.17.0")
SERVER_VERSION_MINIMUM_REQUIRED = version.parse("1.22.0")

INVALID_SERVER_VERSION_MESSAGE = (
    "Server version ({0}) is below the minimum version required "
    f"({SERVER_VERSION_MINIMUM_REQUIRED})"
)


class Client:  # pylint: disable=too-few-public-methods
    """Define a client for the linkding API."""

    def __init__(
        self, url: str, token: str, *, session: ClientSession | None = None
    ) -> None:
        """Initialize.

        Args:
        ----
            url: The full URL to a linkding instance.
            token: A linkding API token.
            session: An optional aiohttp ClientSession.

        """
        self._session = session
        self._token = token
        self._url = url

        self.bookmarks = BookmarkManager(self.async_request)
        self.tags = TagManager(self.async_request)
        self.user = UserManager(self.async_request)

    async def async_request(
        self, method: str, endpoint: str, **kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        """Make an API request.

        Args:
        ----
            method: An HTTP method.
            endpoint: A relative API endpoint.
            **kwargs: Additional kwargs to send with the request.

        Returns:
        -------
            An API response payload.

        Raises:
        ------
            InvalidTokenError: Raised upon an invalid API token.
            RequestError: Raised upon an underlying HTTP error.
            UnknownEndpointError: Raised when requesting an unknown API endpoint.

        """
        kwargs.setdefault("headers", {})
        kwargs["headers"]["Authorization"] = f"Token {self._token}"

        if use_running_session := self._session and not self._session.closed:
            session = self._session
        else:
            session = ClientSession(
                timeout=ClientTimeout(total=DEFAULT_REQUEST_TIMEOUT)
            )

        data: dict[str, Any] = {}

        try:
            async with session.request(
                method, f"{self._url}{endpoint}", **kwargs
            ) as resp:
                data = await resp.json()
                resp.raise_for_status()
        except ClientResponseError as err:
            if resp.status == HTTPStatus.NO_CONTENT:
                # An HTTP 204 will not return parsable JSON data, but it's still a
                # successful response, so we swallow the exception and return:
                return {}
            if resp.status == HTTPStatus.UNAUTHORIZED:
                msg = "Invalid API token"
                raise InvalidTokenError(msg) from err
            if resp.status == HTTPStatus.NOT_FOUND:
                # We break out this particular response for the health check; if we
                # catch this when querying GET /health, we can raise a better final
                # exception than what would normally occur:
                msg = f"Unknown API endpoint: {endpoint}"
                raise UnknownEndpointError(msg) from err
            msg = f"Error while requesting {endpoint}: {data}"
            raise RequestError(msg) from err
        finally:
            if not use_running_session:
                await session.close()

        LOGGER.debug("Data received for %s: %s", endpoint, data)

        return data


async def async_get_client(
    url: str, token: str, *, session: ClientSession | None = None
) -> Client:
    """Get an authenticated, version-checked client.

    Args:
    ----
        url: The full URL to a linkding instance.
        token: A linkding API token.
        session: An optional aiohttp ClientSession.

    Returns:
    -------
        A Client object.

    Raises:
    ------
        InvalidServerVersionError: Raised when the server version is too low.

    """
    client = Client(url, token, session=session)

    try:
        health_resp = await client.async_request("get", "/health")
    except UnknownEndpointError as err:
        raise InvalidServerVersionError(
            INVALID_SERVER_VERSION_MESSAGE.format(
                f"older than {SERVER_VERSION_HEALTH_CHECK_INTRODUCED}"
            )
        ) from err

    server_version = version.parse(health_resp["version"])

    if server_version < SERVER_VERSION_MINIMUM_REQUIRED:
        raise InvalidServerVersionError(
            INVALID_SERVER_VERSION_MESSAGE.format(server_version)
        )

    return client



"""Define package exceptions."""


class LinkDingError(Exception):
    """Define a base exception."""


class InvalidServerVersionError(LinkDingError):
    """Define an error related to an invalid server version."""


class InvalidTokenError(LinkDingError):
    """Define an error related to an invalid API token."""


class RequestError(LinkDingError):
    """An error related to invalid requests."""


class UnknownEndpointError(RequestError):
    """An error related to an unknown endpoint."""



"""Define an object to manage bookmark-based API requests."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any

from aiolinkding.util import generate_api_payload


class BookmarkManager:
    """Define the API manager object."""

    def __init__(self, async_request: Callable[..., Awaitable[dict[str, Any]]]) -> None:
        """Initialize.

        Args:
        ----
            async_request: The request method from the Client object.

        """
        self._async_request = async_request

    async def _async_get_bookmarks(
        self,
        *,
        archived: bool = False,
        query: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        """Return all bookmarks.

        Args:
        ----
            archived: Include archived bookmarks.
            query: Return bookmarks matching a query string.
            limit: Limit the number of returned bookmarks.
            offset: The index at which to return results.

        Returns:
        -------
            An API response payload.

        """
        params = generate_api_payload(
            (
                ("q", query),
                ("limit", limit),
                ("offset", offset),
            )
        )

        endpoint = "/api/bookmarks/"
        if archived:
            endpoint += "archived/"

        return await self._async_request("get", endpoint, params=params)

    async def async_archive(self, bookmark_id: int) -> None:
        """Archive a bookmark.

        Args:
        ----
            bookmark_id: The ID of the bookmark to archive.

        """
        await self._async_request("post", f"/api/bookmarks/{bookmark_id}/archive/")

    async def async_delete(self, bookmark_id: int) -> None:
        """Delete a bookmark.

        Args:
        ----
            bookmark_id: The ID of the bookmark to delete.

        """
        await self._async_request("delete", f"/api/bookmarks/{bookmark_id}/")

    async def async_get_all(
        self,
        *,
        query: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        """Return all bookmarks.

        Args:
        ----
            query: Return bookmarks matching a query string.
            limit: Limit the number of returned bookmarks.
            offset: The index at which to return results.

        Returns:
        -------
            An API response payload.

        """
        return await self._async_get_bookmarks(query=query, limit=limit, offset=offset)

    async def async_get_archived(
        self,
        *,
        query: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        """Return all archived bookmarks.

        Args:
        ----
            query: Return bookmarks matching a query string.
            limit: Limit the number of returned bookmarks.
            offset: The index at which to return results.

        Returns:
        -------
            An API response payload.

        """
        return await self._async_get_bookmarks(
            archived=True, query=query, limit=limit, offset=offset
        )

    async def async_create(
        self,
        url: str,
        *,
        title: str | None = None,
        description: str | None = None,
        notes: str | None = None,
        tag_names: list[str] | None = None,
        is_archived: bool = False,
        unread: bool = False,
        shared: bool = False,
    ) -> dict[str, Any]:
        """Create a new bookmark.

        Args:
        ----
            url: The bookmark URL.
            title: The bookmark title.
            description: The bookmark description.
            notes: Any Markdown-formatted notes.
            tag_names: A list of strings to use as tags.
            is_archived: Immediately archive the bookmark.
            unread: Immediately mark the bookmark as unread.
            shared: Immediately mark the bookmark as shared.

        Returns:
        -------
            An API response payload.

        """
        payload = generate_api_payload(
            (
                ("url", url),
                ("title", title),
                ("description", description),
                ("notes", notes),
                ("tag_names", tag_names),
                ("is_archived", is_archived),
                ("unread", unread),
                ("shared", shared),
            )
        )

        return await self._async_request("post", "/api/bookmarks/", json=payload)

    async def async_get_single(self, bookmark_id: int) -> dict[str, Any]:
        """Return a single bookmark.

        Args:
        ----
            bookmark_id: The ID of the bookmark to get.

        Returns:
        -------
            An API response payload.

        """
        return await self._async_request("get", f"/api/bookmarks/{bookmark_id}/")

    async def async_unarchive(self, bookmark_id: int) -> None:
        """Unarchive a bookmark.

        Args:
        ----
            bookmark_id: The ID of the bookmark to unarchive.

        """
        await self._async_request("post", f"/api/bookmarks/{bookmark_id}/unarchive/")

    async def async_update(
        self,
        bookmark_id: int,
        *,
        url: str | None = None,
        title: str | None = None,
        description: str | None = None,
        notes: str | None = None,
        tag_names: list[str] | None = None,
        unread: bool = False,
        shared: bool = False,
    ) -> dict[str, Any]:
        """Update an existing bookmark.

        Args:
        ----
            bookmark_id: The ID of the bookmark to update.
            url: The bookmark URL.
            title: The bookmark title.
            description: The bookmark description.
            notes: Any Markdown-formatted notes.
            tag_names: A list of strings to use as tags.
            unread: Immediately mark the bookmark as unread.
            shared: Immediately mark the bookmark as shared.

        Returns:
        -------
            An API response payload.

        """
        payload = generate_api_payload(
            (
                ("url", url),
                ("title", title),
                ("description", description),
                ("notes", notes),
                ("tag_names", tag_names),
                ("unread", unread),
                ("shared", shared),
            )
        )

        return await self._async_request(
            "patch", f"/api/bookmarks/{bookmark_id}/", json=payload
        )
